use std::{
    io::Read,
    sync::{Arc, RwLock},
    fs::File,
    path::Path,
};

use trie::Trie;

use crate::util::string::minimize_whitespace;

#[derive(Clone)]
pub struct CacheData {
    pub mime:String,
    pub data:Vec<u8>,
}

struct WebCacheData {
    objects:Trie<CacheData>,
}

#[derive(Clone)]
pub struct WebCache {
    data:Arc<RwLock<WebCacheData>>,
}
impl WebCache {
    pub fn new() -> Self
    {
        Self {
            data:Arc::new(RwLock::new(WebCacheData {
                objects:Trie::new(),
            })),
        }
    }

    pub fn fetch(&self, object:&str) -> Option<CacheData>
    {
        match self.data.read() {
            Ok(reader) => {
                match reader.objects.get(object.as_bytes()) {
                    Some(data) => Some(data.clone()),
                    None => None,
                }
            }
            Err(_) => None,
        }
    }

    pub fn cache(&self, mime:&str, object:&str, data:Vec<u8>)
    {
        match self.data.write() {
            Ok(mut writer) => {
                writer.objects.set(object.as_bytes(), CacheData {
                    mime:mime.to_string(),
                    data,
                });
            },
            Err(_) => { },
        }
    }

    pub fn cache_file<P :AsRef<Path>>(&self, mime:&str, object:&str, path:P) -> Result<(),()>
    {
        match self.data.write() {
            Ok(mut writer) => {
                let mut data = Vec::new();
                if let Ok(mut file) = File::open(path) {
                    file.read_to_end(&mut data).ok();
                    writer.objects.set(object.as_bytes(), CacheData {
                        mime:mime.to_string(),
                        data,
                    });
                    Ok(())
                } else { Err(()) }
            }
            Err(_) => Err(()),
        }
    }

    pub fn cache_whitespace_minimize(&self, object:&str) -> Result<(),()>
    {
        match self.data.write() {
            Ok(mut writer) => {
                if let Some(data) = writer.objects.get_mut(object.as_bytes()) {
                    data.data = minimize_whitespace(&String::from_utf8(data.data.clone()).unwrap()).as_bytes().to_vec();
                    Ok(())
                } else { Err(()) }
            },
            Err(_) => Err(()),
        }
    }

    pub fn cache_file_group<P :AsRef<Path>>(&self, mime:&str, object:&str, paths:&[P]) -> Result<(),()>
    {
        match self.data.write() {
            Ok(mut writer) => {
                let data = paths.into_iter().map(|path| {
                    let mut buffer = Vec::new();
                    if let Ok(mut file) = File::open(path) {
                        file.read_to_end(&mut buffer).ok();
                    }
                    buffer
                }).collect::<Vec<Vec<u8>>>().concat();
                writer.objects.set(object.as_bytes(), CacheData {
                    mime:mime.to_string(),
                    data:data,
                });
                Ok(())
            },
            Err(_) => Err(()),
        }
    }
}
